<?
/*
$experiments = 1;
$filename = "TEXTCZ1.txt";
*/
include("common.php");

if (!isset($experiments)) $experiments = 10; // number of random-experiments to be done on each level
$topten = 10; // number of words to show in "top-ten"
$divisors = array(10, 20, 100, 1000, 10000, 100000); // 1/likelihoods of messing-up

echo "Filename:\t$filename\n";
echo "Experiments:\t$experiments\n";
print_time("Started");

$file = file($filename);
$lines = count($file);
print_time("File loaded");
$topten = min($topten, $lines);

$line0 = " ";
$unigram = array();
$bigram = array();
foreach ($file as $line) {
	$line = trim($line);
	$bigram[$line0][$line]++;
	$unigram[$line]++;
	$line0 = $line;
}
unset($bigram[" "]);
$chars = 0;
$all_chars = array();
foreach ($unigram as $line => $count) {
	for ($i=strlen($line); $i--; ) $all_chars[substr($line, $i, 1)] += $count;
	$chars += $count * strlen($line);
}
$all_chars_keys = array_keys($all_chars); // all characters in number-indexed array
$unigram_keys = array_keys($unigram); // all words in number-indexed array
print_time("Unigrams and bigrams made");

function count_topten($topten, &$unigram) {
	if (!$topten) return 0;
	$in_topten = array();
	foreach ($unigram as $word => $count) {
		if (count($in_topten) < $topten) {
			$in_topten[$word] = $count;
			if (count($in_topten) == $topten) $min = min($in_topten);
		}
		elseif ($count > $min) {
			foreach ($in_topten as $key => $val) if ($val == $min) { unset($in_topten[$key]); break; }
			$in_topten[$word] = $count;
			$min = min($in_topten);
		}
	}
	$return = 0;
	foreach($in_topten as $val) $return += $val;
	return $return;
}

function count_freq(&$unigram, $avg) {
	$return = 0;
	foreach ($unigram as $count) {
		if ($count <= $avg) $return++;
	}
	return $return;
}

echo "\nBasic characteristics:\n";
echo "Number of words in file:\t$lines\n";
echo "Number of characters in file (without \\n):\t$chars\n";
echo "Number of unique characters:\t".count($all_chars)."\n";
printf("Average length of word:\t%1.2f\n", $chars / $lines);

function output($filename, &$all_chars, &$unigram, &$bigram) {
	$fw = fopen($filename, "wb");
	fwrite($fw, "char\tcount\n");
	foreach ($all_chars as $word1 => $count) {
		fwrite($fw, "$word1\t$count\n");
	}
	fwrite($fw, "\nword\tcount\n");
	foreach ($unigram as $word1 => $count) {
		fwrite($fw, "$word1\t$count\n");
	}
	fwrite($fw, "\nword1\tword2\tcount(word1,word2)\tcount(word1)\n");
	foreach ($bigram as $word1 => $line) foreach ($line as $word2 => $count) {
		fwrite($fw, sprintf("$word1\t$word2\t%d\t%d\n", $count, $unigram[$word1]));
	}
	fclose($fw);
}

//output("$filename.dat", $all_chars, $unigram, $bigram); print_time("Unigrams and bigrams saved");

function entropy(&$unigram, &$bigram, $last_word) { // computes conditional entropy
	global $lines, $filename;
	$entropy = 0;
	foreach ($bigram as $word1 => $line) {
		$total = $unigram[$word1];
		if ($word1 === $last_word) $total--;
		if ($total == 0) echo ">! division by zero: $word1 (last word in file: $last_word)\n";
		else foreach ($line as $word2 => $count) $entropy -= $count * log2($count / $total);
	}
	return $entropy / ($lines - 1);
}

function print_entropy_perplex($entropy) { // prints conditional entropy and perplexity, also can print min, max, and avg values
	$format = "%1.8f / %1.4f";
	if (!is_array($entropy)) printf("Conditional entropy H(J|I) / Perplexity PX(J|I):\t$format\n", $entropy, pow(2, $entropy));
	else foreach ($entropy as $key => $val) printf("Conditional entropy H(J|I) / Perplexity PX(J|I): $key:\t$format\n", $val, pow(2, $val));
}

function print_stats(&$unigram, $unique_chars) {
	global $topten, $lines;
	echo "Number of unique words:\t".count($unigram)."\n";
	echo "Frequency of $topten most frequent words:\t".count_topten($topten, $unigram)."\n";
	echo "Number of words with frequency equal to 1:\t".count_freq($unigram, 1)."\n";
	echo "Number of words with frequency less than avg:\t".count_freq($unigram, $lines / count($unigram))."\n";
}

echo "\nOriginal file:\n";
print_stats($unigram, count($all_chars));
print_entropy_perplex(entropy($unigram, $bigram, trim($file[count($file)-1])));
print_time();

if ($experiments) foreach (array("Chars", "Words") as $messing) {
	foreach ($divisors as $divisor) {
		printf("\n%s messed up with a likelihood of %2.3f%%:\n", $messing, 100/$divisor);
		$sum = $max = 0; $min = -1;
		for ($exp=1; $exp <= $experiments; $exp++) {
			mt_srand(getmicrotime() * 100000);
			$line0 = " ";
			$unigram = array();
			$bigram = array();
			foreach ($file as $line) {
				$line = trim($line);
				if ($messing == "Words" && mt_rand(1, $divisor) == 1) {
					$line = $unigram_keys[mt_rand(0, count($unigram_keys) - 1)];
				}
				if ($messing == "Chars") for ($i=0; $i < strlen($line); $i++) if (mt_rand(1, $divisor) == 1) {
					$line[$i] = $all_chars_keys[mt_rand(0, count($all_chars_keys) - 1)];
				}
				$bigram[$line0][$line]++;
				$unigram[$line]++;
				$line0 = $line;
			}
			unset($bigram[" "]);
			$entropy = entropy($unigram, $bigram, $line0);
			$sum += $entropy;
			$min = ($min == -1 ? $entropy : min($min, $entropy));
			$max = max($max, $entropy);
			print_time("After $exp experiment(s)");
		}
		// $all_chars = array(); foreach ($unigram as $line => $count) for ($i=strlen($line); $i--; ) $all_chars[substr($line, $i, 1)] += $count; // count number of unique chars
		print_stats($unigram, count($all_chars)); // stats for last experiment
		print_entropy_perplex($experiments == 1 ? $entropy : array("avg" => $sum / $experiments, "min" => $min, "max" => $max));
		print_time();
	}
}
?>
